@@ -6,11 +6,12 @@
 
     :copyright: (c) 2012 by Matt Wright.
     :copyright: (c) 2017 by CERN.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
 import inspect
+import typing as t
 
 from flask import Markup, current_app, request
 from flask_login import current_user
@@ -27,7 +28,12 @@ from wtforms import (
     ValidationError,
     validators,
 )
-from wtforms.fields.html5 import EmailField
+
+try:  # pragma: no cover
+    # Version 3+
+    from wtforms.fields import EmailField
+except ImportError:
+    from wtforms.fields.html5 import EmailField
 from wtforms.validators import StopValidation
 
 from .babel import is_lazy_string, make_lazy_string
@@ -35,7 +41,7 @@ from .confirmable import requires_confirmation
 from .utils import (
     _,
     _datastore,
-    config_value,
+    config_value as cv,
     do_flash,
     find_user,
     get_identity_attribute,
@@ -73,9 +79,19 @@ _default_field_labels = {
     "identity": _("Identity"),
     "sendcode": _("Send Code"),
     "passcode": _("Passcode"),
+    "username": _("Username"),
 }
 
 
+class NullableStringField(StringField):
+    # Note that current WTForms (2.2.1) has a bug where StringFields can never be
+    # None - it changes them to an empty string. DBs don't like that if you have
+    # your column be 'nullable' and 'unique'.
+    def process_formdata(self, valuelist):
+        if valuelist:
+            self.data = valuelist[0]
+
+
 class ValidatorMixin:
     """
     This is called at import time - so there is no app context.
@@ -99,9 +115,9 @@ class ValidatorMixin:
             not is_lazy_string(self.message) and not self.message
         ):
             # Creat on first usage within app context.
-            cv = config_value("MSG_" + self._original_message)
-            if cv:
-                self.message = make_lazy_string(_local_xlate, cv[0])
+            msg = cv("MSG_" + self._original_message, strict=False)
+            if msg:
+                self.message = make_lazy_string(_local_xlate, msg[0])
             else:
                 self.message = self._original_message
         return super().__call__(form, field)
@@ -170,6 +186,30 @@ def unique_user_email(form, field):
         raise ValidationError(msg)
 
 
+def unique_username(form, field):
+    if not field.data:
+        # N.B. username is NULLABLE in the DB. Whether it is required for an app
+        # is dictated by the USERNAME_REQUIRED config.
+        return
+    uia_username = get_identity_attribute("username")
+    norm_username = _security._username_util.normalize(field.data)
+    if (
+        _datastore.find_user(
+            case_insensitive=uia_username.get("case_insensitive", False),
+            username=norm_username,
+        )
+        is not None
+    ):
+        msg = get_message("USERNAME_ALREADY_ASSOCIATED", username=norm_username)[0]
+        raise ValidationError(msg)
+
+
+def valid_username(form, field):
+    msg, field.data = _security._username_util.validate(field.data)
+    if msg:
+        raise ValidationError(msg)
+
+
 def unique_identity_attribute(form, field):
     """A validator that checks the field data against all configured
     SECURITY_USER_IDENTITY_ATTRIBUTES.
@@ -183,7 +223,7 @@ def unique_identity_attribute(form, field):
     :return: Nothing; if field data corresponds to an existing User, ValidationError
         is raised.
     """
-    for mapping in config_value("USER_IDENTITY_ATTRIBUTES"):
+    for mapping in cv("USER_IDENTITY_ATTRIBUTES"):
         attr = list(mapping.keys())[0]
         details = mapping[attr]
         idata = details["mapper"](field.data)
@@ -268,9 +308,19 @@ class NextFormMixin:
             raise ValidationError(get_message("INVALID_REDIRECT")[0])
 
 
+register_username_field = NullableStringField(
+    get_form_field_label("username"),
+    validators=[valid_username, unique_username],
+)
+
+
 class RegisterFormMixin:
     submit = SubmitField(get_form_field_label("register"))
 
+    # The "username" field is added in init_app if USERNAME_ENABLE is set.
+    # This is just a type hint.
+    username: t.ClassVar[Field]
+
     def to_dict(self, only_user):
         """
         Return form data as dictionary
@@ -352,10 +402,20 @@ class PasswordlessLoginForm(Form, UserEmailFormMixin):
         return True
 
 
+login_email_field = EmailField(
+    get_form_field_label("email"), validators=[email_required]
+)
+
+login_string_field = StringField(
+    get_form_field_label("email"), validators=[email_required]
+)
+
+
 class LoginForm(Form, NextFormMixin):
     """The default login form"""
 
-    email = EmailField(get_form_field_label("email"), validators=[email_required])
+    # Note: "email" field is added at init_app time - depending on whether
+    # username is enabled or not (EmailField versus StringField)
     password = PasswordField(
         get_form_field_label("password"), validators=[password_required]
     )
@@ -366,7 +426,7 @@ class LoginForm(Form, NextFormMixin):
         super().__init__(*args, **kwargs)
         if not self.next.data:
             self.next.data = request.args.get("next", "")
-        self.remember.default = config_value("DEFAULT_REMEMBER_ME")
+        self.remember.default = cv("DEFAULT_REMEMBER_ME")
         if (
             current_app.extensions["security"].recoverable
             and not self.password.description
@@ -447,7 +507,7 @@ class ConfirmRegisterForm(Form, RegisterFormMixin, UniqueEmailFormMixin):
             return False
 
         # To support unified sign in - we permit registering with no password.
-        if not config_value("UNIFIED_SIGNIN"):
+        if not cv("UNIFIED_SIGNIN"):
             # password required
             if not self.password.data or not self.password.data.strip():
                 self.password.errors.append(get_message("PASSWORD_NOT_PROVIDED")[0])
@@ -488,7 +548,7 @@ class RegisterForm(ConfirmRegisterForm, NextFormMixin):
     def validate(self):
         if not super().validate():
             return False
-        if not config_value("UNIFIED_SIGNIN"):
+        if not cv("UNIFIED_SIGNIN"):
             # password_confirm required
             if not self.password_confirm.data or not self.password_confirm.data.strip():
                 self.password_confirm.errors.append(
@@ -583,17 +643,17 @@ class TwoFactorSetupForm(Form, UserEmailFormMixin):
         # TODO: the super class validate is never called - thus we have to
         # initialize errors to lists below. It also means that 'email' is never
         # validated - though it isn't required so the mixin might not be correct.
-        choices = config_value("TWO_FACTOR_ENABLED_METHODS")
+        choices = cv("TWO_FACTOR_ENABLED_METHODS")
         if "email" in choices:
             # backwards compat
             choices.append("mail")
-        if not config_value("TWO_FACTOR_REQUIRED"):
+        if not cv("TWO_FACTOR_REQUIRED"):
             choices.append("disable")
         if "setup" not in self.data or self.data["setup"] not in choices:
             self.setup.errors = list()
             self.setup.errors.append(get_message("TWO_FACTOR_METHOD_NOT_AVAILABLE")[0])
             return False
-        if self.setup.data == "sms" and len(self.phone.data) > 0:
+        if self.setup.data == "sms" and self.phone.data and len(self.phone.data) > 0:
             # Somewhat bizarre - but this isn't required the first time around
             # when they select "sms". Then they get a field to fill out with
             # phone number, then Submit again.
@@ -621,11 +681,11 @@ class TwoFactorVerifyCodeForm(Form, UserEmailFormMixin):
             self.primary_method == "google_authenticator"
             or self.primary_method == "authenticator"
         ):
-            self.window = config_value("TWO_FACTOR_AUTHENTICATOR_VALIDITY")
+            self.window = cv("TWO_FACTOR_AUTHENTICATOR_VALIDITY")
         elif self.primary_method == "email" or self.primary_method == "mail":
-            self.window = config_value("TWO_FACTOR_MAIL_VALIDITY")
+            self.window = cv("TWO_FACTOR_MAIL_VALIDITY")
         elif self.primary_method == "sms":
-            self.window = config_value("TWO_FACTOR_SMS_VALIDITY")
+            self.window = cv("TWO_FACTOR_SMS_VALIDITY")
         else:
             return False
 
@@ -645,7 +705,7 @@ class TwoFactorVerifyCodeForm(Form, UserEmailFormMixin):
 
 
 class TwoFactorRescueForm(Form):
-    """The Two-factor Rescue validation form """
+    """The Two-factor Rescue validation form"""
 
     help_setup = RadioField(
         "Trouble Accessing Your Account?",
